// GDM12864H.c

/*
// Mega644P
// External Crystal, 18.432MHz

CrystalFontz CFAG12864B series Graphics Display serial backpack.
Compatible with the Xiamen Ocular GDM12864H graphics display,
sold by Sparkfun Electronics.

For use with the GDM12864H display. CS1 & CS2 need to be interchanged
Move: solder bridge J7 --> J5
Move: solder bridge J6 --> J8
*/

/*************************************************************/
/*************************************************************/
#include <avr/io.h>
#include <stdio.h>

#include "Driver.h"
#include "ASCII.h"

// System Messages
const char BANNER_MSG_1[] = "XXXXXXXX";
const char BANNER_MSG_2[] = "XXXXXXXXXXXXXXXXX";
const char BANNER_MSG_3[] = "XXXXXXXXXX";

// Global structure to control the graphics LCD display
// These variables are global to maintain the integrity of LCD_PutChar
// and also be able to provide controlled cursor movement
struct {
//      unsigned char Status;
//      unsigned char Cursor;
      unsigned char PageCount;
      unsigned char ColumnPointer;
      unsigned char ColumnData;
      unsigned char ChipSelect;
}LCD;

int LCD_SleepCounter;

/****************************************************************************/

/****************************************************************************/
void LCD_DrawLine (unsigned char x1, unsigned char x2, unsigned char y1, unsigned char y2) {
    char x, y;
    double m;

    x = x1;
    m = (double)(y2 - y1) / (x2 - x1);
    y = (char)((-m*x)+y2);

    while(x < x2) {
       y = (char)((-m*x)+y2);
      LCD_SetPixle (x++, y);
    }
}
/****************************************************************************/

/****************************************************************************/
void LCD_DrawBox (unsigned char x1, unsigned char x2,
               unsigned char y1, unsigned char y2) {
    char n;

    n = x1;
    do {
        LCD_SetPixle (n, y1);
    } while(++n <= x2);

    n = y1;
    do {
        LCD_SetPixle (x2, n);
    } while(++n <= y2);

    n = x2;
    do {
        LCD_SetPixle (n, y2);
    } while(--n >= x1);

    n = y2;
    do {
        LCD_SetPixle (x1, n);
    } while (--n >= y1);
}
/****************************************************************************/

/****************************************************************************/
void LCD_ClearPixle (unsigned char x, unsigned char y) {
    unsigned char d, q, p, n;

    q = y / 8;
    p = q * 8;
    n = y - p;

    d = LCD_GetColumn (x, q);
    LCD.ChipSelect = LCD_PutCursor (x, q);
    LCD_PutData (~(1<<n)&d, LCD.ChipSelect);
}
/****************************************************************************/

/****************************************************************************/
void LCD_SetPixle (unsigned char x, unsigned char y) {
    unsigned char d, q, p, n;

    q = y / 8;
    p = q * 8;
    n = y - p;

    d = LCD_GetColumn (x, q);
    LCD.ChipSelect = LCD_PutCursor (x, q);
    LCD_PutData ((1<<n)|d, LCD.ChipSelect);
}
/****************************************************************************/

/****************************************************************************/
unsigned char LCD_GetColumn (unsigned char x, unsigned char y) {

    LCD.ChipSelect = LCD_PutCursor (x, y);
    LCD_BusyWait(LCD.ChipSelect);
    LCD_DATA_DDR = 0x00; // Set LCD data port to inputs

    LCD_CONTROL_OUT |= (1<<LCD_RW);
    LCD_CONTROL_OUT &= ~(1<<LCD.ChipSelect);
    // There must be a design flaw in the controller so, do a dummy read
    LCD_CONTROL_OUT |= (1<<LCD_E);
    LCD_Delay(10);
    LCD.ColumnData = LCD_DATA_IN; // Dummy read
    LCD_CONTROL_OUT &= ~(1<<LCD_E);
    LCD_Delay(10);
    // Do the correct read of the column data
    LCD_CONTROL_OUT |= (1<<LCD_E);
    LCD_Delay(10);
    LCD.ColumnData = LCD_DATA_IN;
    LCD_CONTROL_OUT &= ~(1<<LCD_E);
    LCD_CONTROL_OUT |= (1<<LCD.ChipSelect);

    return (LCD.ColumnData);
}
/****************************************************************************/

/****************************************************************************/
void LCD_PutColumn (unsigned char character, unsigned char x, unsigned char y) {
    LCD.ChipSelect = LCD_PutCursor (x, y);
    LCD_PutData (character, LCD.ChipSelect);
}
/****************************************************************************/

/****************************************************************************/
unsigned char LCD_PutCursor (unsigned char x, unsigned char y) {
    unsigned char ChipSelect;

    LCD.PageCount = y;
    LCD.ColumnPointer = x;
    if (LCD.ColumnPointer < 64) {
       ChipSelect = LCD_CS1; // Set display left side
      // Set collunm address
      LCD_WrtCmd (COLUMN_ADDRESS | LCD.ColumnPointer, ChipSelect);
    }
    else {
         ChipSelect = LCD_CS2; // Set display right side
        // Set collunm address
        LCD_WrtCmd (COLUMN_ADDRESS | (LCD.ColumnPointer - 64), ChipSelect);
    }
    // Set page address
    LCD_WrtCmd (PAGE_ADDRESS | LCD.PageCount, ChipSelect);
    return (ChipSelect);
}
/****************************************************************************/

/****************************************************************************/
void LCD_PutString (const char *character, unsigned char x, unsigned char y) {
    unsigned char q = 0;

    LCD.ChipSelect = LCD_PutCursor (x, y);
    while (character[q] != NULL) {
          // Check that the next character fits on the current line
         LCD_PutChar (character[q++]);
    }
}
/****************************************************************************/

/****************************************************************************/
char LCD_PutChar (unsigned char character) {
   unsigned char ColumnCounter = 0;

   // Check that the next character fits on the current line
   if ((LCD.ColumnPointer + 6) > 127) {
      LCD.PageCount++;
      LCD.ColumnPointer = 0;
      LCD_CONTROL_OUT |= (1<<LCD_CS2); // Disable the Right side
      LCD_WrtCmd ((PAGE_ADDRESS + LCD.PageCount), LCD_CS1); // Set 'ColumnCounter' page address
      LCD_WrtCmd (COLUMN_ADDRESS, LCD_CS1); // Set 'Y' collunm address
   }

   // Put the next character out onto the display
   for (ColumnCounter = 0; (ColumnCounter < 6); ColumnCounter++) {
      // Check for the end of the current line
      if (LCD.ColumnPointer > 127) {
         LCD.PageCount++;
         LCD_CONTROL_OUT |= (1<<LCD_CS2); // Disable the Right side
         LCD_WrtCmd ((PAGE_ADDRESS + LCD.PageCount), LCD_CS1); // Set page address
         LCD_WrtCmd (COLUMN_ADDRESS, LCD_CS1); // Set 'Y' collunm address
         LCD.ColumnPointer = 0;
      }

      // Check for the last page
      if (LCD.PageCount > 7) {
         LCD_CONTROL_OUT |= (1<<LCD_CS2); // Disable the Right side
         LCD_WrtCmd (PAGE_ADDRESS, LCD_CS1); // Set page address
         LCD_WrtCmd (COLUMN_ADDRESS, LCD_CS1); // Set 'Y' collunm address
         LCD.PageCount = 0;
         LCD.ColumnPointer = 0;
      }

      // Check to see if the Left line is full, if so, select the Right line
      if (LCD.ColumnPointer == 64) {
         LCD_CONTROL_OUT |= (1<<LCD_CS1); // Disable the Left side
         LCD_WrtCmd ((PAGE_ADDRESS + LCD.PageCount), LCD_CS2); // Set page address
         LCD_WrtCmd (COLUMN_ADDRESS, LCD_CS2); // Set 'Y' collunm address
      }

      // Check if Left or Right side should be selected
      if (LCD.ColumnPointer > 63) LCD.ChipSelect = LCD_CS2;
      else LCD.ChipSelect = LCD_CS1;
      // Send the current column to the display
      LCD_PutData (Ascii[character][ColumnCounter], LCD.ChipSelect);
      LCD.ColumnPointer++;
   }
   return(-1);
}
/****************************************************************************/

/****************************************************************************/
void LCD_ClearScreen (void) {
    LCD_ClearPage (LCD_CS1);
    LCD_ClearPage (LCD_CS2);
}
/****************************************************************************/

/****************************************************************************/
void LCD_ClearPage (unsigned char chipselect) {
   unsigned char X, Y;

   X = PAGE_ADDRESS;
   LCD_BusyWait(chipselect);
   do {
         LCD_WrtCmd (X, chipselect); // Set 'X' page address
      LCD_WrtCmd (COLUMN_ADDRESS, chipselect); // Set 'Y' collunm address
         for (Y = 0; (Y < 64); Y++) {
         LCD_PutData (0x00, chipselect); // Set data
      }
   } while (++X < PAGE_ADDRESS + 8);
}
/****************************************************************************/

/****************************************************************************/
// Clock cycle = 67nS @ 14.7456MHz
// Delay resolution ~ 1uS @ 14.7456MHz
void LCD_Delay (unsigned int d) {
    while (d-- != 0);
}
/****************************************************************************/

/****************************************************************************/
void LCD_PutData (unsigned char character, unsigned char chipselect) {
   LCD_BusyWait(chipselect);
   LCD_CONTROL_OUT &= ~((1<<chipselect) | (1<<LCD_RW));

   LCD_CONTROL_OUT |= (1<<LCD_E);
   LCD_DATA_OUT = character;
   LCD_Delay(1);
   LCD_CONTROL_OUT &= ~(1<<LCD_E);

   LCD_CONTROL_OUT |= (1<<chipselect) | (1<<LCD_RW);
}
/****************************************************************************/

/****************************************************************************/
void LCD_WrtCmd (unsigned char cmd, unsigned char chipselect) {
    LCD_BusyWait(chipselect);
   LCD_CONTROL_OUT &= ~((1<<chipselect) | (1<<LCD_RW) | (1<<LCD_DI));

   LCD_CONTROL_OUT |= (1<<LCD_E);
   LCD_DATA_OUT = cmd;
   LCD_Delay(1);
   LCD_CONTROL_OUT &= ~(1<<LCD_E);

   LCD_CONTROL_OUT |= (1<<chipselect) | (1<<LCD_DI) | (1<<LCD_RW);
}
/****************************************************************************/

/****************************************************************************/
void LCD_BusyWait (unsigned char chipselect) {
    char s;
   LCD_DATA_DDR = 0x00; // Set LCD data port to inputs

   LCD_CONTROL_OUT &= ~((1<<LCD_DI) | (1<<chipselect));
   LCD_CONTROL_OUT |= (1<<LCD_RW);
   do {
      LCD_CONTROL_OUT |= (1<<LCD_E);
      LCD_Delay(1);
      s = LCD_DATA_IN;
      LCD_CONTROL_OUT &= ~(1<<LCD_E);
   } while (s & (1<<LCD_BUSY));
   LCD_CONTROL_OUT |= (1<<LCD_DI) | (1<<chipselect);
   LCD_DATA_DDR |= 0xFF; // Set LCD data port to outputs
}
/****************************************************************************/

/****************************************************************************/
void LCD_Init (void) {
   // Initialize the AVR controller I/O
   LCD_DATA_DDR = 0xFF; // Set LCD_DATA_DDR to all outputs
   LCD_DATA_OUT = 0x00; // Set LCD_DATA_OUT to logic low
   LCD_CONTROL_DDR = 0xFF; // Set LCD_CONTROL_DDR to all outputs
   LCD_CONTROL_OUT |= (1<<LCD_DI) | (1<<LCD_RES)
               | (1<<LCD_CS1) | (1<<LCD_CS2);
   LCD_WrtCmd (LCD_ON, LCD_CS1); // Power up the Left side controller of the display
   LCD_WrtCmd (LCD_ON, LCD_CS2); // Power up the Right side controller of the display
   LCD_Sleep(100); // Wait 1 milisecond for the display to stableize
   LCD_ClearScreen ();
}
/****************************************************************************/

/****************************************************************************/
// Debounce any I/O pin (singularly or individually) on any I/O Port
unsigned char SwitchPressed(unsigned char Port, unsigned char Bit) {
         static unsigned char LastKey; // Remember LastKey state, from call to call
         unsigned char DebounceCount = 0;

         do  {
             if (LastKey != (Port & Bit))
                DebounceCount = 10; // Can be as low as 10 counts @ 14.7456MHz

             LastKey = (Port & Bit);
         } while (DebounceCount-- != 0);

         return (~LastKey); // Negated for the STK500/STK600
}
/****************************************************************************/

/****************************************************************************/
// LCD delay timer
// Interrupt once every milisecond
// External Crystal, 18.432MHz
//TIMER1 initialize - prescale: 64
// WGM: 4) CTC, TOP = OCR1A
// Desired resolution: 1.000mS
// Actual resolution:  0.996528mS (0.3742% Error)
// Maximum delay: 65,535 milliseconds
void LCD_Sleep(unsigned int time) {
     LCD_SleepCounter = time;
     while(LCD_SleepCounter > 0);
}
/****************************************************************************/

/****************************************************************************/
// Timer 1 Intrrupt Service Routine
// Interrupt once every milisecond
// External Crystal, 18.432MHz
//TIMER1 initialize - prescale: 64
// WGM: 4) CTC, TOP = OCR1A
// Desired resolution: 1.000mS
// Actual resolution:  0.996528mS (0.3742% Error)
// Maximum delay: 65,535 milliseconds
#pragma interrupt_handler timer1:14
void timer1(void) {
    // Decrement the LCD_SleepCounter
    if (LCD_SleepCounter != 0) LCD_SleepCounter--;
}
/****************************************************************************/

/****************************************************************************/
// Timer 1 Intrrupt Service Routine
// Interrupt once every milisecond
// External Crystal, 18.432MHz
//TIMER1 initialize - prescale: 64
// WGM: 4) CTC, TOP = OCR1A
// Desired resolution: 1.000mS
// Actual resolution:  0.996528mS (0.3742% Error)
void timer1_init(void) {
    TCCR1B = 0x00;      // Stop Timer Counter 1
    TCNT1 = 0xff1B;   // Clear Timer Counter 1
    OCR1A = 0x011F;   // We want a 1.000 Milisecond delay resolution
    TCCR1A = 0x00;      // Generate automatic output
    TCCR1B = (1<<WGM12) | (1<<CS11) | (1<<CS10); // Start Timer OCR1A, CLK / 64
    }
/****************************************************************************/

/****************************************************************************/
// Call this routine to initialize all peripherals
void Init_Devices(void) {
   // Stop errant interrupts until set up
   asm("cli");


   timer1_init();
   MCUCR = 0x00;
   TIMSK |= (1<<OCIE1A);
   asm("sei");
   LCD_Init();
 // All peripherals are now initialized
//    LED_ON();
   LCD_PutString (BANNER_MSG_1, 30, LINE_2); // Send a message to the LCD display
   LCD_PutString (BANNER_MSG_2, 35, LINE_3); // Send a message to the LCD display
   LCD_PutString (BANNER_MSG_3, 1, LINE_5); // Send a message to the LCD display

   LCD_Sleep(50000); // Delay for the banner message to be visible
   LCD_ClearScreen ();
   //LED_OFF();
}
/****************************************************************************/
